/*------------------------------------------------------------------------------

Copyright 2012 University of Aveiro, Portugal, All Rights Reserved.

These programs are supplied free of charge for research purposes only,
and may not be sold or incorporated into any commercial product. There is
ABSOLUTELY NO WARRANTY of any sort, nor any undertaking that they are
fit for ANY PURPOSE WHATSOEVER. Use them at your own risk. If you do
happen to find a bug, or have modifications to suggest, please report
the same to Armando J. Pinho, ap@ua.pt. The copyright notice above
and this statement of conditions must remain an integral part of each
and every copy made of these files.

------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "defs.h"
#include "mem.h"
#include "seqs.h"
#include "context.h"

/*----------------------------------------------------------------------------*/

static HCCounter *zeroCounters;

static void InitHashTable(CModel *cModel)

	{ 
	if((cModel->hTable.entries = (Entry **)Calloc(cModel->hTable.size,
	  sizeof(Entry *))) == NULL)
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if((cModel->hTable.entrySize = (unsigned short *)Calloc(cModel->hTable.size,
	  sizeof(unsigned short))) == NULL)
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if((zeroCounters = (HCCounter *)Calloc(cModel->nSymbols,
	  sizeof(HCCounter))) == NULL)
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	cModel->hTable.nUsedEntries = 0;
	cModel->hTable.nUsedKeys = 0;
	}

/*----------------------------------------------------------------------------*/

static void InitArray(CModel *cModel)

	{
	if(!(cModel->array.counters = (ACCounter *)Calloc(cModel->nPModels *
	  cModel->nSymbols, sizeof(ACCounter))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	}

/*----------------------------------------------------------------------------*/

static void InsertKey(HashTable *hTable, unsigned hIndex, unsigned key,
  unsigned nSymbols)

	{
	if(!(hTable->entries[hIndex] = (Entry *)Realloc(hTable->entries[hIndex],
	  (hTable->entrySize[hIndex] + 1) * sizeof(Entry), sizeof(Entry))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if(!hTable->entrySize[hIndex])
		hTable->nUsedEntries++;

	if(!(hTable->entries[hIndex][hTable->entrySize[hIndex]].counters =
	  (HCCounter *)Calloc(nSymbols, sizeof(HCCounter))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	hTable->nUsedKeys++;
	hTable->entries[hIndex][hTable->entrySize[hIndex]].key = key;
	hTable->entrySize[hIndex]++;
	}

/*----------------------------------------------------------------------------*/

static HCCounter *GetHCCounters(HashTable *hTable, unsigned key)

	{
	unsigned n;
	unsigned hIndex = key % hTable->size; /* The hash index */

	for(n = 0 ; n < hTable->entrySize[hIndex] ; n++)
		if(hTable->entries[hIndex][n].key == key) /* If key found */
			return hTable->entries[hIndex][n].counters;

	return NULL;
	}

/*----------------------------------------------------------------------------*/

PModel *CreatePModel(unsigned nSymbols)

	{
	PModel *pModel;

	if(!(pModel = (PModel *)Malloc(sizeof(PModel))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if(!(pModel->freqs = (unsigned *)Malloc(nSymbols * sizeof(unsigned))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	return pModel;
	}

/*----------------------------------------------------------------------------*/

FloatPModel *CreateFloatPModel(unsigned nSymbols)

	{
	FloatPModel *floatPModel;

	if(!(floatPModel = (FloatPModel *)Malloc(sizeof(FloatPModel))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if(!(floatPModel->freqs = (double *)Malloc(nSymbols * sizeof(double))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	return floatPModel;
	}

/*----------------------------------------------------------------------------*/

void UpdateCModelCounter(CModel *cModel, unsigned pModelIdx, unsigned symbol)

	{
	unsigned n;
	ACCounter *aCounters;

	if(cModel->mode == HASH_TABLE_MODE)
		{
		unsigned i;
		unsigned hIndex = pModelIdx % cModel->hTable.size; /* The hash index */

		for(n = 0 ; n < cModel->hTable.entrySize[hIndex] ; n++)
			{
			/* If key found */
			if(cModel->hTable.entries[hIndex][n].key == pModelIdx)
				{
				if(++cModel->hTable.entries[hIndex][n].counters[symbol] == 255)
					for(i = 0 ; i < cModel->nSymbols ; i++)
						cModel->hTable.entries[hIndex][n].counters[i] >>= 1;

				return;
				}
				
			}

		/* If key not found */
		InsertKey(&cModel->hTable, hIndex, pModelIdx, cModel->nSymbols);
		cModel->hTable.entries[hIndex][n].counters[symbol]++;
		}

	else
		{
		aCounters = &cModel->array.counters[pModelIdx * cModel->nSymbols];
		aCounters[symbol]++;
		if(aCounters[symbol] == cModel->maxCount && cModel->maxCount != 0)
			for(n = 0 ; n < cModel->nSymbols ; n++)
				aCounters[n] >>= 1;

		}

	}

/*----------------------------------------------------------------------------*/

CModel *CreateCModel(unsigned maxCtxSize, unsigned nSymbols, unsigned
  nCtxSymbols, unsigned deltaNum, unsigned deltaDen, unsigned maxCount,
  unsigned hSize, double gamma, char ir)

	{
	CModel *cModel;

	if(!(cModel = (CModel *)Calloc(1, sizeof(CModel))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if(maxCtxSize > 16)
		{
		fprintf(stderr, "Error: context size cannot be greater than 16\n");
		exit(1);
		}

	cModel->nPModels = (ULL)pow(nCtxSymbols, maxCtxSize);
	cModel->maxCtxSize = maxCtxSize;
	cModel->ctxSize = maxCtxSize;
	cModel->nSymbols = nSymbols;
	cModel->nCtxSymbols = nCtxSymbols;
	cModel->deltaNum = deltaNum;
	cModel->deltaDen = deltaDen;
	cModel->gamma = gamma;
	cModel->ir = ir;
	cModel->hTable.size = hSize;
	if((ULL)(cModel->nPModels) * nSymbols * sizeof(ACCounter) >> 20 >
	  MAX_ARRAY_MEMORY)
		{
		cModel->mode = HASH_TABLE_MODE;
		cModel->maxCount = maxCount >> 8;
		InitHashTable(cModel);
		}

	else
		{
		cModel->mode = ARRAY_MODE;
		cModel->maxCount = maxCount;
		InitArray(cModel);
		}

	return cModel;
	}

/*----------------------------------------------------------------------------*/

void ComputePModel(CModel *cModel, PModel *pModel, unsigned pModelIdx)

	{
	int symbol;
	ACCounter *aCounters;
	HCCounter *hCounters;

	pModel->sum = 0;
	if(cModel->mode == HASH_TABLE_MODE)
		{
		if(!(hCounters = GetHCCounters(&cModel->hTable, pModelIdx)))
			hCounters = zeroCounters;

		for(symbol = 0 ; symbol < cModel->nSymbols ; symbol++)
			{
			pModel->freqs[symbol] = cModel->deltaNum + cModel->deltaDen *
			  hCounters[symbol];
			pModel->sum += pModel->freqs[symbol];
			}
		}

	else
		{
		aCounters = &cModel->array.counters[pModelIdx * cModel->nSymbols];
		for(symbol = 0 ; symbol < cModel->nSymbols ; symbol++)
			{
			pModel->freqs[symbol] = cModel->deltaNum + cModel->deltaDen *
			  aCounters[symbol];
			pModel->sum += pModel->freqs[symbol];
			}
		}

	}

/*----------------------------------------------------------------------------*/

double PModelSymbolNats(PModel *pModel, unsigned symbol)

	{
	return log((double)pModel->sum / pModel->freqs[symbol]);
	}

/*----------------------------------------------------------------------------*/

double FractionOfPModelsUsed(CModel *cModel)

	{
	unsigned pModel, symbol, sum, counter = 0;
	ACCounter *aCounters;
	HCCounter *hCounters;

	sum = 0;
	for(pModel = 0 ; pModel < cModel->nPModels ; pModel++)
		{
		if(cModel->mode == HASH_TABLE_MODE)
			{
			hCounters = GetHCCounters(&(cModel->hTable), pModel);
			if(hCounters)
				counter++;

			}

		else
			{
			aCounters = &(cModel->array.counters[pModel * cModel->nSymbols]);
			for(symbol = 0 ; symbol < cModel->nSymbols ; symbol++)
				sum += aCounters[symbol];

			if(sum != 0)
				counter++;

			}

		}

	return (double)counter / cModel->nPModels;
	}

/*----------------------------------------------------------------------------*/

double FractionOfPModelsUsedOnce(CModel *cModel)

	{
	unsigned pModelIdx;
	unsigned symbol, sum, counter = 0;
	ACCounter *aCounters;
	HCCounter *hCounters;

	sum = 0;
	for(pModelIdx = 0 ; pModelIdx < cModel->nPModels ; pModelIdx++)
		{
		if(cModel->mode == HASH_TABLE_MODE)
			{
			hCounters = GetHCCounters(&(cModel->hTable), pModelIdx);

			if(!hCounters)
				continue;

			for(symbol = 0 ; symbol < cModel->nSymbols ; symbol++)
				sum += hCounters[symbol];

			if(sum == 1)
				counter++;

			}

		else
			{
			aCounters = &(cModel->array.counters[pModelIdx * cModel->nSymbols]);
			for(symbol = 0 ; symbol < cModel->nSymbols ; symbol++)
				sum += aCounters[symbol];

			if(sum == 1)
				counter++;

			}

		}

	return (double)counter / cModel->nPModels;
	}

/*----------------------------------------------------------------------------*/

void HashingStats(CModel *cModel)

	{
	unsigned entry, n, k, emptyEntries = 0, nSmallCounters = 0,
	  maxEntrySize = 0;
	ULL possibleKeys;
	double average = (double)cModel->hTable.nUsedKeys / cModel->hTable.size;
	double deviation = 0;

	for(entry = 0 ; entry < cModel->hTable.size ; entry++)
		{
		deviation += fabs(average - cModel->hTable.entrySize[entry]);

		if(!cModel->hTable.entrySize[entry])
			{
			emptyEntries++;
			continue;
			}

		if(cModel->hTable.entrySize[entry] > maxEntrySize)
			maxEntrySize = cModel->hTable.entrySize[entry];

		k = 0;
		for(n = 0 ; n < cModel->hTable.entrySize[entry] ; n++)
			{
			/*
			 * If "counters" is non-zero, then this key did not require
			 * "large" counters
			 */
			if(cModel->hTable.entries[entry][n].counters != 0)
				nSmallCounters++;

			/*
			 * Keeps counting the number of HCCounters in this entry.
			 * counters = 0 means that the corresponding key has been
			 * generated more than once and, therefore, the set of
			 * counters has been allocated.
			 */
			if(cModel->hTable.entries[entry][n].counters == 0)
				k++;

			}

		}

	possibleKeys = powl((double)cModel->nSymbols, (double)cModel->maxCtxSize);

	printf("Hash size ......... %u\n", cModel->hTable.size);
	printf("Used entries ...... %u\n", cModel->hTable.nUsedEntries);
	printf("Ideal entry size .. %.1f\n", average);
	printf("Deviation ......... %.1f\n", deviation / cModel->hTable.size);
	printf("Used keys ......... %u [%.2f %% of %llu]\n",
	  cModel->hTable.nUsedKeys, 100.0 * (double)cModel->hTable.nUsedKeys /
	  possibleKeys, possibleKeys);
	printf("Small counters .... %u\n", nSmallCounters);
	printf("Large counters .... %u\n", cModel->hTable.nUsedKeys-nSmallCounters);
	printf("Max entry size .... %u\n", maxEntrySize);
	}

/*----------------------------------------------------------------------------*/

void ResetCModel(CModel **cModel)

	{
	unsigned entry, n;

	if((*cModel)->mode == ARRAY_MODE)
		memset((*cModel)->array.counters, 0, (*cModel)->nPModels *
		  (*cModel)->nSymbols * sizeof(ACCounter));

	else
		{
		for(entry = 0 ; entry < (*cModel)->hTable.size ; entry++)
			{
			for(n = 0 ; n < (*cModel)->hTable.entrySize[entry] ; n++)
				Free((*cModel)->hTable.entries[entry][n].counters,
				  ((*cModel)->nSymbols * sizeof(HCCounter)));

			if((*cModel)->hTable.entrySize[entry])
				{
				Free((*cModel)->hTable.entries[entry],
				  ((*cModel)->hTable.entrySize[entry] * sizeof(Entry)));
				(*cModel)->hTable.entries[entry] = NULL;
				(*cModel)->hTable.entrySize[entry] = 0;
				}

			}

		(*cModel)->hTable.nUsedEntries = 0;
		(*cModel)->hTable.nUsedKeys = 0;
		}

    }

/*------------------------------------------------------------------------------

	"symbol" points to the symbol being encoded / decoded

------------------------------------------------------------------------------*/

unsigned GetPModelIdx(Base *base, CTemplate *cTemplate, Alphabet *alphabet)

	{
	unsigned n, idx = 0;

	for(n = 0 ; n < cTemplate->size ; n++)
		idx += alphabet->ctxSymbols[*(base - cTemplate->position[n])] << (n<<1);

	return idx;
	}

/*------------------------------------------------------------------------------

	"symbol" points to the symbol being encoded / decoded

------------------------------------------------------------------------------*/

unsigned GetPModelIdxC(Base *base, CTemplate *cTemplate, Alphabet *alphabet)

	{
	unsigned n, idx = 0;

	base++;
	for(n = 0 ; n < cTemplate->size ; n++)
		idx += (3 - alphabet->ctxSymbols[*(base - cTemplate->position[n])]) <<
		  ((cTemplate->size - n - 1) << 1);

	return idx;
	}

/*----------------------------------------------------------------------------*/

void IncreaseTemplateStorage(CTemplate *cTemplate, int additionalSize)

	{
	if(cTemplate->size == 0)
		{
		if(!(cTemplate->position = (int *)Malloc(additionalSize * sizeof(int))))
			{
			fprintf(stderr, "Error: in memory allocation\n");
			exit(1);
			}

		}

	else
		{
		if(!(cTemplate->position = (int *)Realloc(cTemplate->position,
		  (additionalSize + cTemplate->size) * sizeof(int), additionalSize *
		  sizeof(int))))
			{
			fprintf(stderr, "Error: in memory allocation\n");
			exit(1);
			}

		}

	}

/*------------------------------------------------------------------------------

	Pow function from http://martin.ankerl.com/2007/10/04/
	  optimized-pow-approximation-for-java-and-c-c/

------------------------------------------------------------------------------*/

double Pow(double a, double b)
	{
	int tmp = (*(1 + (int *)&a));
	int tmp2 = (int)(b * (tmp - 1072632447) + 1072632447);
	double p = 0.0;
	*(1 + (int * )&p) = tmp2;
	return p;
	}

