/*------------------------------------------------------------------------------

Copyright 2012 University of Aveiro, Portugal, All Rights Reserved.

These programs are supplied free of charge for research purposes only,
and may not be sold or incorporated into any commercial product. There is
ABSOLUTELY NO WARRANTY of any sort, nor any undertaking that they are
fit for ANY PURPOSE WHATSOEVER. Use them at your own risk. If you do
happen to find a bug, or have modifications to suggest, please report
the same to Armando J. Pinho, ap@ua.pt. The copyright notice above
and this statement of conditions must remain an integral part of each
and every copy made of these files.

------------------------------------------------------------------------------*/

/*
 * Data structures for handling finite-context models.
 */

#ifndef CONTEXT_H_INCLUDED
#define CONTEXT_H_INCLUDED

#define N_SYMBOLS					4

#define DEFAULT_PMODEL_DELTA_NUM	1
#define DEFAULT_PMODEL_DELTA_DEN	1

#define MAX_ARRAY_MEMORY			1024 /* DNA: size 12 = 128 MB */
#define ARRAY_MODE					0
#define HASH_TABLE_MODE				1

typedef unsigned short ACCounter; /* Size of context counters for arrays */
typedef unsigned char  HCCounter; /* Size of context counters for hash tables */

#define PMODEL_MAX_COUNT			((1 << (sizeof(ACCounter) * 8)) - 1)
/* If 0, never divide counters by maxCount */

typedef struct
	{
	unsigned		key;		/* The context index associated to this entry */
	HCCounter		*counters;	/* The context counters */
	}
Entry;

typedef struct
	{
	unsigned		size;		/* Size of the hash table */
	unsigned short	*entrySize;	/* Number of keys in this entry */
	Entry			**entries;	/* The heads of the hash table lists */
	unsigned		nUsedEntries;
	unsigned		nUsedKeys;
	}
HashTable;

typedef struct
	{
	ACCounter	*counters;
	}
Array;

typedef struct
	{
	unsigned	*freqs;
	unsigned	sum;
	}
PModel;

typedef struct
	{
	double		*freqs;
	}
FloatPModel;

typedef struct
	{
	unsigned	maxCtxSize;	/* Maximum depth of context template */
	unsigned	ctxSize;	/* Current depth of context template */
	unsigned	nSymbols;	/* Number of coding symbols */
	unsigned	nCtxSymbols;/* Number of symbols used for context computation */
	ULL			nPModels;	/* Maximum number of probability models */
	unsigned	deltaNum;	/* Numerator of delta */
	unsigned	deltaDen;	/* Denominator of delta */
	unsigned	maxCount;	/* Counters /= 2 if one counter >= maxCount */
	double		gamma;		/* Performance decaying parameter */
	char		ir;			/* If 'y', update the inverted complements */
	unsigned	mode;
	HashTable	hTable;
	Array		array;
	}
CModel;

typedef struct
	{
	int size;
	int *position;
	int deepestPosition;
	}
CTemplate;

PModel *CreatePModel(unsigned nSymbols);
FloatPModel *CreateFloatPModel(unsigned nSymbols);
void UpdateCModelCounter(CModel *cModel, unsigned pModelIdx, unsigned symbol);
CModel *CreateCModel(unsigned maxCtxSize, unsigned nSymbols, unsigned
  nCtxSymbols, unsigned deltaNum, unsigned deltaDen, unsigned maxCount,
  unsigned hSize, double gamma, char ir);
double FractionOfPModelsUsed(CModel *cModel);
double FractionOfPModelsUsedOnce(CModel *cModel);
void ComputePModel(CModel *cModel, PModel *pModel, unsigned pModelIdx);
double PModelSymbolNats(PModel *pModel, unsigned symbol);
void HashingStats(CModel *cModel);
void ResetCModel(CModel **cModel);
unsigned GetPModelIdx(Base *bases, CTemplate *cTemplate, Alphabet *alphabet);
unsigned GetPModelIdxC(Base *bases, CTemplate *cTemplate, Alphabet *alphabet);
void IncreaseTemplateStorage(CTemplate *cTemplate, int additionalSize);
double Pow(double a, double b);

#endif /* CONTEXT_H_INCLUDED */

